// HP80DIS - Everett Kaser - January 2015
/*********************  Header Files  *********************/

#include <windows.h>
#include <stdio.h>
//#include <stdlib.h>
#include <io.h>
#include <fcntl.h>
#include <time.h>
#include <string.h>
//#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>

// DEVTIME sets default command line values when debugging in the development environment.
// Set it to 0 before release.
#define	DEVTIME	0

/********************* Global Variables *******************/
char			*TitleRevStr = "HP80DIS 2.0 (19 Feb 2015)";

BOOL			BailOut = FALSE;	// in case an unknown error occurs processing the inputs

#define	PATHLEN	512
char			BinFileName[PATHLEN]="";
char			ListFileName[PATHLEN];

BOOL 			Is85=TRUE;	// default to HP-85 unless -87 option is specified
BOOL			OnlySrc=FALSE;	// defaults to OFF unless -s option is specified
BOOL			IsROM=TRUE; // default to ROM unless .BPGM file extension on input file
BOOL			ForceComment=FALSE;	// if add a '!' on lines that don't have it

unsigned char	*InBuf;		// ptr to malloc'd input ROM or BPGM bytes
long			InBufLen;

char			*LstBuf;	// ptr to malloc'd input LST file
char			*pLST;		// working ptr into LstBuf
long			LstBufLen;

int				OutFile;	// file handle to which all output is written (temp .LST file)

char			TmpFileName[]="_tmp.tmp.LST";

char			eolseq[]="\r\n";
int				eollen=2;

char			ScratBuf[8192];

char			*pLstLines[65536];	// ptrs to each line of the input .LST file
									// up to 8 pointers for EVERY byte of the ROM.  Should be enough.
int				LstLinesCnt;

//**********************
// Disassembly stuff
//

#define	GLO_DAD		0
#define	GLO_EQU		1

struct gloval {
	char	type;
	char	label[11];// really only use 9 max, the other 2 keeps the structure a multiple of 4 in len
	DWORD	address;
};
typedef struct gloval GLOBAL;

GLOBAL	Globals[8192];
int		GlobalsCnt;

long	BaseAddress, EndAddress;
BOOL	IsSystemROM=FALSE;

#define	REC_COD	0
#define	REC_BYT	1
#define	REC_ASC	2
#define	REC_ASP	3
#define	REC_BSZ	4
#define	REC_DEF	5
#define	REC_VAL	6
#define	REC_RAM	7
#define	REC_I_O	8
#define	REC_DRP	9
#define	REC_REM	10
#define	REC_UND	11
#define	REC_DAD	12
#define	REC_EQU 13

#define	FLG_ROMJSB_DEF	1
#define	FLG_FDA			2
#define	FLG_DAD_DEFINED	4
#define	FLG_FEQ			8
#define	FLG_EQU_DEFINED 16
#define	FLG_GLO			32
#define	FLG_GTO			64

struct codrec {
	BYTE			type;
	BYTE			flags;
	char			label[9];	// line-label
	char 			reflabel[9];// argument EQU label reference
	WORD			datover;	// over ride of how many data bytes there are for disassembly
	DWORD			address;
		// Note: the 'address' field contains the actual HP-85 address shifted over 16 bits.
		// If the lower 16 bits is 0xFFFF, then this record applies to real bytes in the HP-85.
		// Otherwise (lower 16 bits == 0x0000 to 0xFFFE), this record merely contains a full-line
		// comment that precedes the actual HP-85 bytes.  It's done this way for ease of
		// sorting/finding/positioning full-line comments and HP-85 addresses relative to
		// each other.  This allows up to 65535 lines of comments before any one line of HP-85
		// code.  Should be enough.
	char			*Comment;
	struct codrec	*pNext;
	struct codrec	*pPrev;
};

typedef struct codrec DISASMREC;

#define	FindRecNext(f)	(f->pNext)
#define	FindRecPrev(f)	(f->pPrev)
#define	HP85ADDR(n)		(((n)>>16)&0x0000FFFF)
#define	SkipWhite(p)	while( *(p) && (*(p)==' ' || *(p)=='\t') ) ++(p)

#define	MAXROMS		32

// ptr to blocks of DISASMRECs for code we're disassembling
// For each of the pDis entries, there are two values, the first ([0]) is the pointer to the
// FIRST record, and the second ([1]) is the pointer to the LAST record.
DISASMREC	*pDis[2];

// ROMnums[] contains the ROM numbers of the supported option-ROMs.
// numROMS, of course, keeps track of how many option-ROMs (including ROM0) have been loaded.
int	ROMnums[MAXROMS] = {
	0010,	// Service
	0050, 	// Assembler
	0250,	//
	0260,	//
	0300,	// I/O
	0317,	//
	0320,	// Mass Storage
	0340,	//
	0350,	//
	0360	//
	};
int			numROMS=10, curROM=0, RomPos;
BYTE		DisArp=0xFF, DisDrp=0xFF, DisTyp=REC_COD, LastDisTyp;
WORD		DisDataOverride;
//DWORD		ScrollIndex[2048];
//#define	NUMSI	(sizeof(ScrollIndex)/sizeof(DWORD))

DWORD		NxtByt, NxtRom, saveNxtByt;	// "pointer" to next byte to disassemble
DISASMREC	*NextRec, *pDefLabel=NULL, ScratchDISASM;

char	*opc[256] = {
	// 000
	"ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ",
	"ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ",
	"ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ",
	"ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ",
	"ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ",
	"ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ",
	"ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ",
	"ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ","ARP ",
	// 100
	"DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ",
	"DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ",
	"DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ",
	"DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ",
	"DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ",
	"DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ",
	"DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ",
	"DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ","DRP ",
	// 200
	"ELB ","ELM ","ERB ","ERM ","LLB ","LLM ","LRB ","LRM ",
	"ICB ","ICM ","DCB ","DCM ","TCB ","TCM ","NCB ","NCM ",
	"TSB ","TSM ","CLB ","CLM ","ORB ","ORM ","XRB ","XRM ",
	"BIN ","BCD ","SAD ","DCE ","ICE ","CLE ","RTN ","PAD ",
	"LDB ","LDM ","STB ","STM ","LDBD","LDMD","STBD","STMD",
	"LDB ","LDM ","STB ","STM ","LDBI","LDMI","STBI","STMI",
	"LDBD","LDMD","STBD","STMD","LDBD","LDMD","STBD","STMD",
	"LDBI","LDMI","STBI","STMI","LDBI","LDMI","STBI","STMI",
	// 300
	"CMB ","CMM ","ADB ","ADM ","SBB ","SBM ","JSB ","ANM ",
	"CMB ","CMM ","ADB ","ADM ","SBB ","SBM ","JSB ","ANM ",
	"CMBD","CMMD","ADBD","ADMD","SBBD","SBMD","NOP3","ANMD",
// The above NOP3 (opcode 326) is not a real opcode and was
// never used (so far as I know) by any Series 80 code.  It is
// an officially unsupported opcode in the HP-85 CPU.
// The hardware engineers investigated it for me in the early 1980's
// and found that, the way the CPU was designed, it resulted in a
// 3-byte "no operation".  What it does is fetch the following two
// bytes from memory, but does nothing with them.  So, this code:
//	BYT	326
//	BYT skip1
//	BYT skip2
//	BYT nextopcode
// will skip the skip1 and skip2 bytes and then resume execution
// with the nextopcode byte, with no modifications to the CPU registers,
// status, or memory, other than the incrementing of the PC (reg 4).
// I include it here just for the sake of completeness.
	"CMBD","CMMD","ADBD","ADMD","SBBD","SBMD","NOP1","ANMD",
// The above NOP1 (opcode 336) is not a real opcode and was
// never used (so far as I know) by any Series 80 code.  It is
// an officially unsupported opcode in the HP-85 CPU.
// The hardware engineers investigated it for me in the early 1980's
// and found that, the way the CPU was designed, it resulted in a
// 1-byte "no operation".  So, this code:
//	BYT	336
//	BYT nextopcode
// fetches the 326 opcode, but does nothing, and then resumes execution
// with the nextopcode byte, with no modifications to the CPU registers,
// status, or memory, other than the incrementing of the PC (reg 4).
	"POBD","POMD","POBD","POMD","PUBD","PUMD","PUBD","PUMD",
	"POBI","POMI","POBI","POMI","PUBI","PUMI","PUBI","PUMI",
	"JMP ","JNO ","JOD ","JEV ","JNG ","JPS ","JNZ ","JZR ",
	"JEN ","JEZ ","JNC ","JCY ","JLZ ","JLN ","JRZ ","JRN "
};

DWORD	base=8;
BYTE	HexDig[] = "0123456789ABCDEF";

/***********************************************/
DISASMREC	*FindEQUlabel(char *label) {

	DISASMREC	*psrch;

	psrch = pDis[0];
	for(; psrch && (psrch->type!=REC_EQU || stricmp((char*)psrch->label, label)); psrch = psrch->pNext) ;
	return psrch;
}
/***********************************************/
DISASMREC	*FindNextEQUaddress(DISASMREC *psrch, DWORD val) {

	if( psrch==NULL ) psrch = pDis[0];
	else psrch = psrch->pNext;
	for(; psrch && (psrch->type!=REC_EQU || HP85ADDR(psrch->address)!=val); psrch = psrch->pNext) ;
	return psrch;
}
/***********************************************/
DISASMREC	*FindDADlabel(char *label) {

	DISASMREC	*psrch;

	psrch = pDis[0];
	for(; psrch && (psrch->type!=REC_DAD || stricmp((char *)psrch->label, label)); psrch = psrch->pNext) ;
	return psrch;
}
/***********************************************/
DISASMREC	*FindNextDADaddress(DISASMREC *psrch, DWORD addr) {

	if( psrch==NULL ) psrch = pDis[0];
	else psrch = psrch->pNext;
	for(; psrch && (psrch->type!=REC_DAD || HP85ADDR(psrch->address)!=addr); psrch = psrch->pNext) ;
	return psrch;
}

//************************************************
DISASMREC *AddRec(BYTE typ, BYTE flags, char *lbl, DWORD addr, char *rem, WORD datcnt) {

	DISASMREC	*ptmp, *psrch;

	if( NULL!=(ptmp=(DISASMREC *)malloc(sizeof(DISASMREC))) ) {
		if( pDis[0]==NULL ) {
			pDis[0] = pDis[1] = ptmp;
			ptmp->pNext = ptmp->pPrev = NULL;
		} else {
			for(psrch=pDis[0]; psrch!=NULL && psrch->address <= addr; psrch = psrch->pNext) ;
			if( psrch==NULL ) {
				pDis[1]->pNext = ptmp;
				ptmp->pPrev = pDis[1];
				ptmp->pNext = NULL;
				pDis[1] = ptmp;
			} else {
				ptmp->pNext = psrch;
				if( psrch->pPrev==NULL ) {
					pDis[0] = ptmp;
					ptmp->pPrev = NULL;
				} else {
					ptmp->pPrev = psrch->pPrev;
					psrch->pPrev->pNext = ptmp;
				}
				psrch->pPrev = ptmp;
			}
		}
		ptmp->type = typ;
		ptmp->flags = flags;
		strcpy((char *)ptmp->label, lbl);
		ptmp->reflabel[0] = 0;
		ptmp->address = addr;
		if( rem != NULL && rem[0]!=0 ) {
			ptmp->Comment = (char *)malloc(strlen(rem)+1);
			if( ptmp->Comment ) strcpy(ptmp->Comment, rem);
		} else ptmp->Comment = NULL;
		ptmp->datover = datcnt;
	}
	return ptmp;
}

/**********************************************************/
DISASMREC	*FindRecAddress(DWORD addr) {

	DISASMREC	*psrch;
//	DWORD		addr16;

//	addr16 = HP85ADDR(addr);
	psrch = pDis[0];
	for(; psrch && psrch->address < addr; psrch = psrch->pNext) ;
	if( psrch && psrch->address != addr ) psrch = NULL;
	return psrch;
}

/**********************************************************/
DISASMREC	*FindRecEQU(char *pLabel) {

	DISASMREC	*psrch;
//	DWORD		addr16;

//	addr16 = HP85ADDR(addr);
	psrch = pDis[0];
	for(; psrch && (psrch->type!=REC_EQU || stricmp(psrch->label, pLabel)); psrch = psrch->pNext) ;
	return psrch;
}

/**********************************************************/
DISASMREC	*FindRecDAD(char *pLabel) {

	DISASMREC	*psrch;
//	DWORD		addr16;

//	addr16 = HP85ADDR(addr);
	psrch = pDis[0];
	for(; psrch && (psrch->type!=REC_DAD || stricmp(psrch->label, pLabel)); psrch = psrch->pNext) ;
	return psrch;
}

/*******************************************************/
GLOBAL	*FindGlobal(DWORD addr, BOOL romjsb) {

	int		i;

	if( romjsb || addr<BaseAddress || addr>=EndAddress || IsSystemROM ) {	// if address is outside 60000-77777 range OR we're disassembling the System ROM (option ROM 0), search the GLOBAL table
		for(i=0; i<GlobalsCnt; i++) {
			if( Globals[i].type==GLO_DAD && Globals[i].address==addr ) {
				return &Globals[i];
			}
		}
	}
	return NULL;
}

/*******************************************************/
GLOBAL	*FindGlobalLabel(char *label) {

	int		i;

	for(i=0; i<GlobalsCnt; i++) {
		if( !stricmp(Globals[i].label, label) ) {
			return &Globals[i];
		}
	}
	return NULL;
}

/*******************************************************/
void FormatReg(char *pBuf, BYTE r) {

//	char	tmpbuf[16];

	if( r==0xFF ) strcpy(pBuf, "r#");
	else if( r==0x01 ) strcpy(pBuf, "r*");
	else {
		sprintf(pBuf, "r%o", r);
//		strcpy(pBuf, tmpbuf+1);	// throw away leading '0' (octal is assumed)
	}
}

/*********************************************************/
void FormatAddress(char *pBuf, WORD addr, DISASMREC *pLineRec) {

	DISASMREC	*pRec;
	GLOBAL		*pGlob;
	BYTE		flags;

	flags = (pLineRec==NULL) ? 0 : pLineRec->flags;

if( addr==03572 ) {
	addr = 03572;
}
	if( flags & FLG_ROMJSB_DEF ) {	// it could be in a System ROM (InBuf[NxtRom+2]==0), this ROM (InBuf[NxtRom+2]==curROM), another ROM (InBuf[NxtRom+2]!=curROM)
		if( InBuf[NxtRom+2]==0 && NULL!=(pGlob=FindGlobal(addr, TRUE)) ) strcpy(pBuf, pGlob->label);
		else if( BaseAddress>=060000 && InBuf[NxtRom+2]!=curROM ) {
			do {
				pRec = FindNextDADaddress(NULL, addr);
			} while( pRec!=NULL && pRec->datover!=InBuf[NxtRom+2] );
			if( pRec!=NULL ) strcpy(pBuf, pRec->label);
			else sprintf(pBuf, "%6.6o", addr);
		} else if( NULL==(pRec=FindRecAddress((addr<<16) | 0x0000FFFF)) || pRec->label[0]==0 ) sprintf(pBuf, "%6.6o", addr);
		else strcpy(pBuf, pGlob->label);
	} else if( flags & FLG_FEQ ) {	// we're forced to search EQUs for
		if( pLineRec->reflabel[0] ) {
			pRec = FindRecEQU(pLineRec->reflabel);
			if( pRec && !(pRec->flags & FLG_EQU_DEFINED) ) {
				pRec->flags |= FLG_EQU_DEFINED;
				pRec->address = addr;
			}
			strcpy(pBuf, pLineRec->reflabel);
		} else sprintf(pBuf, "%6.6o", addr);
	} else if( flags & FLG_FDA ) {	// we're forced to search DADs for
		if( pLineRec->reflabel[0] ) {
			pRec = FindRecDAD(pLineRec->reflabel);
			if( pRec && !(pRec->flags & FLG_DAD_DEFINED) ) {
				pRec->flags |= FLG_DAD_DEFINED;
				pRec->address = addr;
			}
			strcpy(pBuf, pLineRec->reflabel);
		} else sprintf(pBuf, "%6.6o", addr);
	} else if( flags & FLG_GLO ) {	// we're forced to search GLOs for
		if( pLineRec->reflabel[0] ) {
			pGlob = FindGlobalLabel(pLineRec->reflabel);
			if( pGlob && pGlob->address==addr ) strcpy(pBuf, pLineRec->reflabel);
			else printf("ERROR: GLOBAL not found: .fgl %6.6o %s", addr, pLineRec->reflabel);
		} else sprintf(pBuf, "%6.6o", addr);
	} else {
//		if( addr>=BaseAddress && addr<EndAddress || addr<077777 ) {
			if( NULL==(pRec=FindRecAddress((addr<<16) | 0x0000FFFF)) || pRec->label[0]==0 ) {
				if( NULL==(pGlob=FindGlobal(addr, FALSE)) ) sprintf(pBuf, "%6.6o", addr);
				else strcpy(pBuf, pGlob->label);
			} else strcpy(pBuf, (char *)pRec->label);
//		} else sprintf(pBuf, "%6.6o", addr);
	}
}

/********************************************************************
 * NxtByt is offset into ROM[] of what to disassemble
 *	(actually, it's ((offset into ROM[])<<16) | c) where 'c' is
 *	0x0000FFFF if we're pointing to actual ROM code, and some
 *	value from 0x00000000 to 0x0000FFFE if we're pointing to a
 *	comment DISASMREC that PRECEDES acutal ROM code.
 * pBuf is pointer to output buffer
 ********************************************************************/
void DisAsm() {

	BYTE	cod1, cod2, cod3, ARPDRP1, ARPDRP2, AD1, AD2;
	char	arpstr[6], drpstr[6], addstr[16], *pForm, *pBuf;
	DWORD	i, k, s, m, w;
	WORD	addr;
	DISASMREC *ThisRec;
	BOOL	haveComment;

	pBuf = ScratBuf;

//	TargetAddress = 0xFFFFFFFF;

	pForm = "%6.6o  %3.3o";

	k = HP85ADDR(NxtByt) - BaseAddress;
	sprintf(pBuf, pForm, BaseAddress + k, InBuf[k]);
	k = strlen(pBuf);
	pBuf += k;
	while( k++ < 24 ) *pBuf++ = ' ';
	s = NxtRom = HP85ADDR(NxtByt) - BaseAddress;	// just 0-0xFFFF address in ROM (disregarding full-line comments)

	// try to find DISASMREC for this "line" (NxtByt address) of code
	if( NextRec==NULL ) NextRec = pDis[0];
	while( NextRec != NULL  &&  (NextRec->address < NxtByt || NextRec->type==REC_DAD || NextRec->type==REC_EQU) ) {
		if( NextRec->type < REC_UND ) DisTyp = NextRec->type;
		NextRec = NextRec->pNext;
	}

	if( NextRec!=NULL							// if found a record
	 && NextRec->address > NxtByt				// but it's HIGHER than the one we're looking for
	 && (NxtByt & 0x0000FFFF)!=0x0000FFFF ) {	// and the one we're looking for is NOT code
		if( NextRec->address <= (NxtByt | 0x0000FFFF) ) {	// if record is for next full-line comment OR NxtByt's "real" code
			NxtByt = NextRec->address;							// then jump to there
		} else NxtByt |= 0x0000FFFF;						// else jump NxtByt to "real" code
	}
	saveNxtByt = NxtByt;

	// check for one byte in between tables of DEFs, and force COD type so next table comes out
	// (this is rare, but when it occurs, it can be ugly in the disassembly)
	if( NextRec!=NULL								// if found a DISASMREC after NxtByt
	 && (NextRec->address & 0x0000FFFF)==0x0000FFFF	// and it points to "real" code
	 && (NxtByt & 0x0000FFFF)==0x0000FFFF			// and NxtByt points to "real" code
	 && NextRec->address==(NxtByt+0x00010000)		// and the following DISASMREC "real" code is 1 byte after the NxtByt "real" code
	 && DisTyp==REC_DEF								// and the previous code was a DEF address
	 && NextRec->type!=REC_DEF )					// and the next code is a DEF address
		DisTyp = REC_COD;					// then use REC_COD for this one byte, not REC_DEF, to keep the following table aligned

	if( DisTyp==REC_REM ) DisTyp = LastDisTyp;
	if( NextRec!=NULL && NextRec->address==NxtByt ) {
		ThisRec = NextRec;
// we got a DISASMREC for this line of code
		if( (NxtByt & 0x0000FFFF)!=0x0000FFFF ) {	// not "real" code, just a full-line comment
			++NxtByt;		// so we move to next line next time
			// erase the address/object code that was output above
			for(pBuf=ScratBuf, k=0; k<24; k++) *pBuf++ = ' ';

			*pBuf++ = '!';	// output full-line comment
			*pBuf++ = ' ';
			if( ThisRec->Comment ) strcpy(pBuf, ThisRec->Comment);
			else *pBuf = 0;
			return;			// all done with this line
		}
		strcpy(pBuf, (char *)ThisRec->label);
		for(cod1=0; cod1<9; cod1++, pBuf++) if( *pBuf==0 ) break;
		for(; cod1<9; cod1++) *pBuf++ = ' ';
		*pBuf = 0;
		if( ThisRec->type != REC_UND ) DisTyp = ThisRec->type;
		// if label (not blanks), set DRP/ARP to undefined
		if( ThisRec->label[0]!=0 ) DisDrp = DisArp = 0xFF;
		DisDataOverride = ThisRec->datover;
	} else {
		ThisRec = NULL;
// no DISASMREC found, output blanks for label
		strcpy(pBuf, "         ");
		pBuf += 9;
		DisDataOverride = 0;
	}
	// output "code", based upon disassembly type
	LastDisTyp = DisTyp;
	switch( DisTyp ) {
	 case REC_ASC:	// ASCII string of "datover" len (max 8 chars)
		if( (i=DisDataOverride)==0 ) i = 1;
		strcpy(pBuf, "ASC  \"");
		pBuf += strlen(pBuf);
		while( i-- && pBuf<(ScratBuf+256-3) ) *pBuf++ = InBuf[NxtRom++];
		*pBuf++ = '"';
		*pBuf = 0;
		break;
	 case REC_ASP:	// ASCII string with parity (most-significant) bit set on last character
		strcpy(pBuf, "ASP  \"");
		pBuf += strlen(pBuf);
		while( !(InBuf[NxtRom] & 128) && pBuf<(ScratBuf+256-3) ) {
			*pBuf++ = InBuf[NxtRom++];
		}
		if( (InBuf[NxtRom] & 0x7F)<32 ) *pBuf++ = ' ';
		else *pBuf++ = InBuf[NxtRom] & 0x7F;
		++NxtRom;
		*pBuf++ = '"';
		*pBuf = 0;
		break;
	 case REC_DEF:	// DEF statement to define an address (as in a table)
	 	addr = (WORD)(((WORD)InBuf[NxtRom]) + (((WORD)InBuf[NxtRom+1])<<8));
	 	FormatAddress(addstr, addr, ThisRec);
		sprintf(pBuf, "DEF  %s", addstr);
		++NxtRom;
		++NxtRom;
		break;
	 case REC_DRP:	// stand-alone DRP or ARP
		cod1 = InBuf[NxtRom];
		AD1 = cod1 & 0xC0;
		ARPDRP1 = cod1 & 0x3F;
		goto disarpdrp;
	 case REC_COD:	// "normal" assembly code (CPU instructions)
		cod1 = InBuf[NxtRom];
		AD1 = cod1 & 0xC0;
		ARPDRP1 = cod1 & 0x3F;
		cod2 = InBuf[NxtRom+1];
		AD2 = cod2 & 0xC0;
		ARPDRP2 = cod2 & 0x3F;
		cod3 = InBuf[NxtRom+2];
		if( AD1 & 0x80 ) {	// no ARP or DRP
			if( cod1==0xC6 ) goto jsbinx;
			else if( cod1==0xCE ) {	// JSB =
				FormatAddress(addstr, (WORD)((WORD)cod2+(((WORD)cod3)<<8)), ThisRec);
				sprintf(pBuf, "%s =%s", opc[cod1], addstr);
				++NxtRom;
				++NxtRom;
				++NxtRom;
				DisArp = DisDrp = 0xFF;
			} else if( ((cod1 & 0xFC)==0x94)
			 || ((cod1 & 0xF8)==0xA0)
			 || ((cod1 & 0xFC)==0xAC)
			 || ((cod1 & 0xFC)==0xB4)
			 || ((cod1 & 0xFC)==0xBC)
			 || ((cod1 & 0xF8)==0xC0)
			 || ((cod1 & 0xF8)==0xD8)
			 || ((cod1 & 0xF0)==0xE0) ) goto regreg;
			else if( ((cod1 & 0xF8)==0x80)
				  || ((cod1 & 0xFC)==0x88)
				  || ((cod1 & 0xFC)==0x8C)
				  || ((cod1 & 0xFC)==0x90) ) goto reg;
			else if( (cod1 & 0xF0)==0xF0 ) {	// jmp opcodes (JNZ, JNC, etc)
				FormatAddress(addstr, (WORD)(BaseAddress + NxtRom+2+(WORD)((int)((char)cod2))), ThisRec);
				sprintf(pBuf, "%s %s", opc[cod1], addstr);
				++NxtRom;
				++NxtRom;
			} else if( (cod1 & 0xF8)==0x98 ) {	// single byte opcodes (BIN,BCD,SAD,DCE,ICE,CLE,RTN,PAD)
				sprintf(pBuf, "%s", opc[cod1]);
				++NxtRom;
				if( cod1==0x9F ) DisArp = DisDrp = 0xFF;	// if PAD, ARP/DRP are unknown
			} else if( ((cod1 & 0xFC)==0xA8) || ((cod1 & 0xF8)==0xC8) ) goto regimm;
			else if( ((cod1 & 0xFC)==0xB0)
				  || ((cod1 & 0xFC)==0xB8)
				  || ((cod1 & 0xF8)==0xD0) ) goto regdir;
			else goto unknown;
		} else if( !(AD1 & 0x80) && !(AD2 & 0x80)	// first 2 bytes are ARP or DRP
		 && ((AD1==0x40 && AD2==0x00) || (AD1==0x00 && AD2==0x40))
		 && (((cod3 & 0xFC)==0x94)
		  || ((cod3 & 0xF8)==0xA0)
		  || ((cod3 & 0xFC)==0xAC)
		  || ((cod3 & 0xFC)==0xB4)
		  || ((cod3 & 0xFC)==0xBC)
		  || ((cod3 & 0xF8)==0xC0)
		  || ((cod3 & 0xF8)==0xD8)
		  || ((cod3 & 0xF0)==0xE0)) ) {
			if( AD1==0x40 ) {
				DisDrp = ARPDRP1;
				DisArp = ARPDRP2;
			} else {
				DisDrp = ARPDRP2;
				DisArp = ARPDRP1;
			}
			++NxtRom;
			++NxtRom;
			cod1 = cod3;
regreg:
			strcpy(pBuf, opc[cod1]);
			FormatReg(drpstr, DisDrp);
			FormatReg(arpstr, DisArp);
			++NxtRom;
			if( (cod1 & 0xF0)==0xE0 ) {
				if( cod1 & 0x02 ) sprintf(pBuf, "%s %s,-%s", opc[cod1], drpstr, arpstr);
				else sprintf(pBuf, "%s %s,+%s", opc[cod1], drpstr, arpstr);
			} else if( (cod1 & 0xFC)==0xB4 || (cod1 & 0xFC)==0xBC ) {
				arpstr[0] = 'X';
				FormatAddress(addstr, (WORD)((WORD)InBuf[NxtRom]+(((WORD)InBuf[NxtRom+1])<<8)), ThisRec);
				++NxtRom;
				++NxtRom;
				sprintf(pBuf, "%s %s,%s,%s", opc[cod1], drpstr, arpstr, addstr);
			} else sprintf(pBuf, "%s %s,%s", opc[cod1], drpstr, arpstr);
		} else if( (!(AD1 & 0x80) && !(AD2 & 0x80)) || (AD1==0x40 && cod2==0xCE) ) {	// stand-alone ARP or DRP
disarpdrp:
			if( AD1==0x40 ) DisDrp = ARPDRP1;
			else DisArp = ARPDRP1;
			FormatReg(drpstr, ARPDRP1);
			sprintf(pBuf, "%s %s", opc[cod1], drpstr);
			++NxtRom;
		} else if( AD1==0x00 && cod2==0xC6 ) {	// jsb xREG,ADDR
			DisArp = ARPDRP1;
			cod1 = cod2;
			cod2 = cod3;
			++NxtRom;
jsbinx:
			++NxtRom;
			++NxtRom;
			FormatReg(arpstr, DisArp);
			arpstr[0] = 'X';
			FormatAddress(addstr, (WORD)(((WORD)cod2+(((WORD)InBuf[NxtRom])<<8))), ThisRec);
			sprintf(pBuf, "%s %s,%s", opc[cod1], arpstr, addstr);
			++NxtRom;
			DisArp = DisDrp = 0xFF;
		} else if( AD1==0x00
		 && (((cod2 & 0xFC)==0x94)
		  || ((cod2 & 0xF8)==0xA0)
		  || ((cod2 & 0xFC)==0xAC)
		  || ((cod2 & 0xFC)==0xB4)
		  || ((cod2 & 0xFC)==0xBC)
		  || ((cod2 & 0xF8)==0xC0)
		  || ((cod2 & 0xF8)==0xD8)
		  || ((cod2 & 0xF0)==0xE0)) ) {
			DisArp = ARPDRP1;
			++NxtRom;
			cod1 = cod2;
			goto regreg;
		} else if( AD1==0x40
		 && (((cod2 & 0xFC)==0x94)
		  || ((cod2 & 0xF8)==0xA0)
		  || ((cod2 & 0xFC)==0xAC)
		  || ((cod2 & 0xFC)==0xB4)
		  || ((cod2 & 0xFC)==0xBC)
		  || ((cod2 & 0xF8)==0xC0)
		  || ((cod2 & 0xF8)==0xD8)
		  || ((cod2 & 0xF0)==0xE0)) ) {
			DisDrp = ARPDRP1;
			++NxtRom;
			cod1 = cod2;
			goto regreg;
		} else if( AD1==0x40
		 && (((cod2 & 0xF8)==0x80)
		  || ((cod2 & 0xFC)==0x88)
		  || ((cod2 & 0xFC)==0x8C)
		  || ((cod2 & 0xFC)==0x90)) ) {		// single register opcodes (ORB, TSB, etc)
			DisDrp = ARPDRP1;
			++NxtRom;
			cod1 = cod2;
reg:
			FormatReg(drpstr, DisDrp);
			sprintf(pBuf, "%s %s", opc[cod1], drpstr);
			++NxtRom;
		} else if( AD1==0x40
		 && (((cod2 & 0xFC)==0xA8)
		  || ((cod2 & 0xF8)==0xC8)) ) {	// literal immediate (LDB=, CMB=, etc)
			DisDrp = ARPDRP1;
			++NxtRom;
			cod1 = cod2;
			cod2 = cod3;
regimm:
			++NxtRom;
			FormatReg(drpstr, DisDrp);
			if( cod1==0251 && (DisDrp==4 || (ThisRec && (ThisRec->flags & FLG_GTO))) ) {	// LDM R4,=   Change it to GTO xxxx
				++NxtRom;	// skip first byte of address which is in cod2
				FormatAddress(addstr, 1 + (WORD)((WORD)cod2+(((WORD)InBuf[NxtRom])<<8)), ThisRec);
				sprintf(pBuf, "GTO %s", addstr);
				++NxtRom;	// skip second byte of address
				DisDrp = DisArp = 0xFF; // DRP & ARP are undefined after a GTO
			} else if( (cod1 & 1) && ((DisDataOverride>1) || (DisDrp<32 && !(DisDrp & 1)) || (DisDrp>=32 && (DisDrp & 7)<7)) ) {	// multibyte
				if( (DisDataOverride==0 && (DisDrp<32 || (DisDrp>=32 && (DisDrp & 7)==6)))
				 || (DisDataOverride==2 && ThisRec && (ThisRec->flags & FLG_FEQ))
				 || (DisDataOverride==2 && ThisRec && (ThisRec->flags & FLG_GLO))
				 || (DisDataOverride==2 && ThisRec && (ThisRec->flags & FLG_FDA))
					) {
					FormatAddress(addstr, (WORD)((WORD)cod2+(((WORD)InBuf[NxtRom+1])<<8)), ThisRec);
					sprintf(pBuf, "%s %s,=%s", opc[cod1], drpstr, addstr);
					++NxtRom;
					++NxtRom;
				} else {
					sprintf(pBuf, "%s %s,=", opc[cod1], drpstr);
					if( DisDataOverride==0 ) cod2 = DisDrp & 7;
					else cod2 = 8-DisDataOverride;
					pForm = "0%o,";
					while( cod2<7 ) {
						pBuf += strlen(pBuf);
						sprintf(pBuf, pForm, (WORD)InBuf[NxtRom]);
						++NxtRom;
						++cod2;
					}
					pBuf += strlen(pBuf);
					pForm = "0%o";
					sprintf(pBuf, pForm, (WORD)InBuf[NxtRom]);
					++NxtRom;
				}
			} else {			// single byte
				pForm = "%s %s,=0%o";
				sprintf(pBuf, pForm, opc[cod1], drpstr, (WORD)cod2);
				++NxtRom;
			}
		} else if( AD1==0x40
		 && (((cod2 & 0xFC)==0xB0)
		  || ((cod2 & 0xFC)==0xB8)
		  || ((cod2 & 0xF8)==0xD0)) ) {
			DisDrp = ARPDRP1;
			cod1 = cod2;
			cod2 = cod3;
			++NxtRom;
regdir:
			++NxtRom;
			++NxtRom;
			FormatReg(drpstr, DisDrp);
			FormatAddress(addstr, (WORD)((WORD)cod2+(((WORD)InBuf[NxtRom])<<8)), ThisRec);
			sprintf(pBuf, "%s %s,=%s", opc[cod1], drpstr, addstr);
			++NxtRom;
		} else {
unknown:
			if( !(cod1 & 0x80) ) goto disarpdrp;
			if( base==8 ) pForm = "BYT  %o";
			else if( base==10 ) pForm = "BYT  %d";
			else if( base==16 ) pForm = "BYT  0x%X";
			sprintf(pBuf, pForm, cod1);
			++NxtRom;
		}
		break;
	 case REC_RAM:
		if( base==8 ) pForm = "RAM  %o";
		else if( base==10 ) pForm = "RAM  %d";
		else if( base==16 ) pForm = "RAM  0x%X";
		i = (DisDataOverride==0)?1:DisDataOverride;
		sprintf(pBuf, pForm, i);
		NxtRom += i;
		break;
	 case REC_I_O:
		if( base==8 ) pForm = "I/O  %o";
		else if( base==10 ) pForm = "I/O  %d";
		else if( base==16 ) pForm = "I/O  0x%X";
		i = (DisDataOverride==0)?1:DisDataOverride;
		sprintf(pBuf, pForm, i);
		NxtRom += i;
		break;
	 case REC_BSZ:
		if( base==8 ) pForm = "BSZ  %o";
		else if( base==10 ) pForm = "BSZ  %d";
		else if( base==16 ) pForm = "BSZ  0x%X";
		i = (DisDataOverride==0)?1:DisDataOverride;
		sprintf(pBuf, pForm, i);
		NxtRom += i;
		break;
	 case REC_BYT:
	 default:
		if( base==8 ) pForm = "%o";
		else if( base==10 ) pForm = "%d";
		else if( base==16 ) pForm = "0x%X";
		if( (i=DisDataOverride)==0 ) i = 1;
		strcpy(pBuf, "BYT  ");
		pBuf += strlen(pBuf);
		while( i-- && pBuf<(ScratBuf+256-3) ) {
			sprintf(pBuf, pForm, InBuf[NxtRom++]);
			pBuf += strlen(pBuf);
			if( i ) *pBuf++ = ',';
		}
		*pBuf = 0;
		break;
	}
	NxtByt = (NxtRom+BaseAddress)<<16;	// leave lower word == 0 in case of full-line comments in front of next "real" code
	// if DISASMREC and comment, output the comment
	haveComment = ThisRec!=NULL && ThisRec->address==saveNxtByt && ThisRec->Comment!=NULL;
	if( ForceComment || haveComment ) {
		k = strlen(ScratBuf);
		pBuf = ScratBuf + k;
		do {
			*pBuf++ = ' ';
		} while( k++ < 50 );
		*pBuf++ = '!';
		*pBuf++ = ' ';
		if( haveComment ) strcpy(pBuf, ThisRec->Comment);
		else *pBuf = 0;
	}

	m = 4;
	w = 4;
	if( base==8 ) pForm = "%3.3o";
	else if( base==10 ) pForm = "%3.3d";
	else {
		pForm = "%2.2X";
		m = 5;
		w = 3;
	}
	for(k=1; k<min(m, NxtRom-s); k++) {
		sprintf(ScratBuf+8+k*w, pForm, InBuf[s+k]);
		ScratBuf[8+k*w+w-1] = ' ';
	}
	if( isdigit(ScratBuf[0]) ) ScratBuf[23] = ':';
//	ScratBuf[maxlen] = 0;
}

/***********************************************
 * Format current timestamp into a char buffer *
 ***********************************************/
char* TimeString() {

	static char		timbuf[24];
	char		ampm[4];
	struct tm	*now=NULL;
	int			hour=0;
	time_t		timval=0;

	timval = time(NULL);
	now = localtime(&timval);
	hour = now->tm_hour;
	if( hour>12 ) {
		hour -= 12;
		strcpy(ampm, "PM");
	} else {
		strcpy(ampm, "AM");
	}
	sprintf(timbuf, "%2.2d/%2.2d/%2.2d %2.2d:%2.2d:%2.2d %s", now->tm_mon+1, now->tm_mday, now->tm_year+1900, hour, now->tm_min, now->tm_sec, ampm);

	return timbuf;
}

/****************************************
 * Write a line to the output LST file	*
 ****************************************/
void WriteLine(char *buf) {

	if( !OnlySrc || *buf!='.' ) {	// if OnlySrc, discard all DOT-COMMANDS from output
		if( OnlySrc ) buf += 24;	// if OnlySrc, discard address, object bytes, and colon
		write(OutFile, buf, strlen(buf));
		write(OutFile, eolseq, eollen);
	}
}

/***************************************
 * get a number (octal or dec) or return 0
 ******************************************/
DWORD	GetNum(char **p) {

	char	*pc, *pe;
	DWORD	num;

	num = 0;
	pc = *p;
	SkipWhite(pc);
	for(pe=pc; isdigit(*pe); pe++);
	if( toupper(*pe)=='D' ) {	// decimal
		num = atol(pc);
		pc = pe+1;
	} else {					// octal
		while( *pc>='0' && *pc<='7' ) num = (num<<3) + (*pc++ - '0');
	}
	*p = pc;
	return num;
}
/*****************************************************
 * Read LstBuf and convert GLOBAL entries to GLOBALS *
 *****************************************************/
void ProcessGLOBALS() {

	int		i, remlen, linenum;//, lastType, dataType, dataCnt;
	char	*bp, label[10];

	// build array of line ptrs into GLOBALS file and replace \r and \n with NUL
	pLST = LstBuf;
	LstLinesCnt = 0;
	while( (remlen = LstBufLen - (pLST-LstBuf)) > 0 ) {
		// save ptr to start of line
		pLstLines[LstLinesCnt++] = pLST;
		// skip to end of line
		while( remlen && *pLST!='\r' && *pLST!='\n' ) {
			++pLST;
			--remlen;
		}
		// skip past end of line sequence and replace it with NUL chars
		if( remlen && *pLST=='\r' ) {
			*pLST++ = 0;
			--remlen;
		}
		if( remlen && *pLST=='\n' ) {
			*pLST++ = 0;
			--remlen;
		}
	}
	// now process the lines, building DISASMRECs for DOT-COMMANDS, LABELS, and COMMENTS
	for(linenum=0; linenum<LstLinesCnt; linenum++) {
		bp = pLstLines[linenum];
		SkipWhite(bp);
		if( *bp && *bp!='*' ) {	// lines starting with '*' are full-line comments, so ignore them
			// Get up to 8 chars of label
			for(i=0; i<8 && *bp && *bp!=' ' && *bp!='\t'; ) label[i++] = *bp++;
			label[i] = 0;
			if( i==(Is85?6:8) && *bp!=' ' && *bp!='\t' ) printf("ERROR: GLOBAL label too long: %s", label);
			// skip WhiteSpace
			SkipWhite(bp);
			if( !strnicmp(bp, "EQU", 3) ) Globals[GlobalsCnt].type = GLO_EQU;
			else if( !strnicmp(bp, "DAD", 3) ) Globals[GlobalsCnt].type = GLO_DAD;
			else {
				printf("ERROR: GLOBAL unknown statement: %s", pLstLines[linenum]);
				continue;
			}
			if( *bp ) ++bp;	// skip DAD or EQU
			if( *bp ) ++bp;	// skip DAD or EQU
			if( *bp ) ++bp;	// skip DAD or EQU
			SkipWhite(bp);
			Globals[GlobalsCnt].address = GetNum(&bp);	// get address or value
			strcpy(Globals[GlobalsCnt].label, label);
			++GlobalsCnt;
		}
	}
}

/****************************************************
 * Read LstBuf and convert some stuff to DISASMRECs *
 ****************************************************/
void ProcessLST() {

	int		i, remlen, cnt, linenum, dadcnt, equcnt;
	char	*bp, label[10], comment[256];
	DWORD	dw, lastAddress;
	DISASMREC	*pRec;

	//======================================================================
	// build array of line ptrs into .LST file and replace \r and \n with NUL
	pLST = LstBuf;
	LstLinesCnt = 0;
	while( (remlen = LstBufLen - (pLST-LstBuf)) > 0 ) {
		// save ptr to start of line
		pLstLines[LstLinesCnt++] = pLST;
		// skip to end of line
		while( remlen && *pLST!='\r' && *pLST!='\n' ) {
			++pLST;
			--remlen;
		}
		// skip past end of line sequence and replace it with NUL chars
		if( remlen && *pLST=='\r' ) {
			*pLST++ = 0;
			--remlen;
		}
		if( remlen && *pLST=='\n' ) {
			*pLST++ = 0;
			--remlen;
		}
	}

	// now process the lines, building DISASMRECs for DOT-COMMANDS, LABELS, and COMMENTS
	lastAddress = 0;	// in case file starts with full-line comment
	dadcnt = equcnt = 0;
	for(linenum=0; linenum<LstLinesCnt; linenum++) {
		bp = pLST = pLstLines[linenum];
		SkipWhite(bp);
		// process line
		if( *bp=='.' ) { // if it's a command
			if( !strnicmp(bp, ".dis", 4) ) {	// discard

			} else if( !strnicmp(bp, ".byt", 4) ) {	// add a REC_BYT record at the specified address
				bp += 5;
				dw = (GetNum(&bp)<<16)|0xFFFF;	// get BYT address
				cnt = GetNum(&bp);	// get optional number of bytes (or default to 0)
				if( NULL==(pRec=FindRecAddress(dw)) ) AddRec(REC_BYT, 0, "", dw, NULL, cnt);
				else {
					pRec->type = REC_BYT;
					pRec->datover = cnt;
				}
			} else if( !strnicmp(bp, ".arp", 4) || !strnicmp(bp, ".drp", 4) ) {	// add a REC_DRP record at the specified address
				bp += 5;
				dw = (GetNum(&bp)<<16)|0xFFFF;	// get DRP address
				if( NULL==(pRec=FindRecAddress(dw)) ) AddRec(REC_DRP, 0, "", dw, NULL, 0);
				else pRec->type = REC_DRP;
			} else if( !strnicmp(bp, ".def", 4) ) {	// add a REC_DEF record at the specified address
				bp += 5;
				dw = (GetNum(&bp)<<16)|0xFFFF;	// get DEF address
				cnt = GetNum(&bp);	// get possible FLG_ROMJSB_DEF flag
				if( NULL==(pRec=FindRecAddress(dw)) ) {
						AddRec(REC_DEF, cnt, "", dw, NULL, 0);
				}else {
					pRec->type = REC_DEF;
					pRec->flags = cnt;
				}
			} else if( !strnicmp(bp, ".cod", 4) ) {	// add a REC_COD record at the specified address
				bp += 5;
				dw = (GetNum(&bp)<<16)|0xFFFF;	// get COD address
				cnt = GetNum(&bp);	// get optional flag (0=normal .cod, 1=force 1-byte stand-alone DRP/ARP instruction)
				if( NULL==(pRec=FindRecAddress(dw)) ) AddRec(REC_COD, 0, "", dw, NULL, cnt);
				else {
					pRec->type = REC_COD;
					pRec->datover = cnt;
				}
			} else if( !strnicmp(bp, ".asc", 4) ) {	// add a REC_ASC record at the specified address
				bp += 5;
				dw = (GetNum(&bp)<<16)|0xFFFF;	// get ASC address
				cnt = GetNum(&bp);	// get len of ASC string
				if( NULL==(pRec=FindRecAddress(dw)) ) AddRec(REC_ASC, 0, "", dw, NULL, cnt);
				else {
					pRec->type = REC_ASC;
					pRec->datover = cnt;
				}
			} else if( !strnicmp(bp, ".asp", 4) ) {	// add a REC_ASP record at the specified address
				bp += 5;
				dw = (GetNum(&bp)<<16)|0xFFFF;	// get ASP address
				if( NULL==(pRec=FindRecAddress(dw)) ) AddRec(REC_ASP, 0, "", dw, NULL, 0);
				else pRec->type = REC_ASP;
			} else if( !strnicmp(bp, ".bsz", 4) ) {	// add a REC_BSZ record at the specified address
				bp += 5;
				dw = (GetNum(&bp)<<16)|0xFFFF;	// get BSZ address
				cnt = GetNum(&bp);	// get len of BSZ block
				if( NULL==(pRec=FindRecAddress(dw)) ) AddRec(REC_BSZ, 0, "", dw, NULL, cnt);
				else {
					pRec->type = REC_BSZ;
					pRec->datover = cnt;
				}
			} else if( !strnicmp(bp, ".lab", 4) ) {	// make sure a REC exists with the LABEL at the specified address
				bp += 5;
				dw = (GetNum(&bp)<<16)|0xFFFF;	// get label address
				SkipWhite(bp);
				for(i=0; i<8 && *bp && *bp!=' ' && *bp!='\t'; ) label[i++] = *bp++;
				label[i] = 0;
				if( NULL==(pRec=FindRecAddress(dw)) ) AddRec(REC_UND, 0, label, dw, NULL, 0);
				else if( pRec->label[0]==0 ) strcpy((char*)pRec->label, label);
			} else if( !strnicmp(bp, ".eol", 4) ) {	// make sure a REC exists with the EOL COMMENT at the specified address
				bp += 5;
				dw = (GetNum(&bp)<<16)|0xFFFF;	// get comment address
				if( *bp==' ' || *bp=='\t' ) ++bp;  // skip ONE whitespace, to allow 'indenting' of eol comments
				for(i=0; i<255 && *bp; ) comment[i++] = *bp++;
				if( i==0 ) comment[i++] = ' ';	// make sure comment is not empty
				comment[i] = 0;
				if( NULL==(pRec=FindRecAddress(dw)) ) AddRec(REC_UND, 0, "", dw, comment, 0);
				else if( pRec->Comment==NULL ) {
					pRec->Comment = (char *)malloc(strlen(comment)+1);
					if( pRec->Comment ) strcpy(pRec->Comment, comment);
				} // ignore the DOT-COMMAND comment if there's already a comment stored for this line
			} else if( !strnicmp(bp, ".com", 4) ) {	// make sure a REC exists with the FULL-LINE COMMENT at the specified address
				bp += 5;
				dw = (GetNum(&bp)<<16)&0xFFFF0000;	// get comment address
				SkipWhite(bp);
				cnt = GetNum(&bp); // sequence# of this full-line comment at this address
				if( *bp==' ' || *bp=='\t' ) ++bp;  // skip ONE whitespace, to allow 'indenting' of comments
				for(i=0; i<255 && *bp; ) comment[i++] = *bp++;
				if( i==0 ) comment[i++] = ' ';	// make sure comment is not empty
				comment[i] = 0;
				if( NULL==(pRec=FindRecAddress(dw)) ) AddRec(REC_REM, 0, "", dw, comment, 0);
				else {
					do {
						if( pRec->datover==cnt ) break; // there's already a comment with this index
						if( pRec->pNext==NULL || ((pRec->pNext)->address & 0xFFFF0000)!=dw || ((pRec->pNext)->address & 0x0000FFFF)>=0x00008000 ) {
							dw = pRec->address + 1;	// put this as the next full-line comment
							if( (pRec->address & 0x00008000)==0 ) AddRec(REC_REM, 0, "", dw, comment, dw & 0x00007FFF); // add the comment as long as there are less than 32K of them.  Heh-heh.
							break;
						}
						pRec = pRec->pNext;
					} while( pRec != NULL );
				}
			} else if( !strnicmp(bp, ".feq", 4) ) {	// make sure a REC_EQU exists with the reflabel
				bp += 5;
				dw = GetNum(&bp);	// get ref address
				SkipWhite(bp);
				for(i=0; i<8 && *bp && *bp!=' ' && *bp!='\t'; ) label[i++] = *bp++;
				label[i] = 0;
				dw = (dw<<16)|0x0000FFFF;
				if( NULL==(pRec=FindRecEQU(label)) ) pRec = AddRec(REC_EQU, 0, label, 0, NULL, 0);	// create the Rec but FLG_EQU_DEFINED not set yet
				if( NULL==(pRec=FindRecAddress(dw)) ) pRec = AddRec(REC_UND, FLG_FEQ, "", dw, NULL, 0);	// find or add a REC_ for this address and set it's FLG_FEQ
				else pRec->flags |= FLG_FEQ;
				if( pRec!=NULL ) strcpy(pRec->reflabel, label);	// if REC_ found or added, copy 'label' to the 'reflabel'
			} else if( !strnicmp(bp, ".fda", 4) ) {	// make sure a REC_DAD exists with the reflabel
				bp += 5;
				dw = GetNum(&bp);	// get ref address
				SkipWhite(bp);
				for(i=0; i<8 && *bp && *bp!=' ' && *bp!='\t'; ) label[i++] = *bp++;
				label[i] = 0;
				dw = (dw<<16)|0x0000FFFF;
				if( NULL==(pRec=FindRecDAD(label)) ) pRec = AddRec(REC_DAD, 0, label, 0, NULL, 0);	// create the Rec but FLG_DAD_DEFINED not set yet
				if( NULL==(pRec=FindRecAddress(dw)) ) pRec = AddRec(REC_UND, FLG_FDA, "", dw, NULL, 0);	// find or add a REC_ for this address and set it's FLG_FDA
				else pRec->flags |= FLG_FDA;
				if( pRec!=NULL ) strcpy(pRec->reflabel, label);	// if REC_ found or added, copy 'label' to the 'reflabel'
			} else if( !strnicmp(bp, ".fgl", 4) ) {	// make sure a REC_GLO exists with the reflabel
				bp += 5;
				dw = GetNum(&bp);	// get ref address
				SkipWhite(bp);
				for(i=0; i<8 && *bp && *bp!=' ' && *bp!='\t'; ) label[i++] = *bp++;
				label[i] = 0;
				dw = (dw<<16)|0x0000FFFF;
				if( NULL==(pRec=FindRecAddress(dw)) ) pRec = AddRec(REC_UND, FLG_GLO, "", dw, NULL, 0);	// find or add a REC_ for this address and set it's FLG_FDA
				else pRec->flags |= FLG_GLO;
				if( pRec!=NULL ) strcpy(pRec->reflabel, label);	// if REC_ found or added, copy 'label' to the 'reflabel'
			} else if( !strnicmp(bp, ".gto", 4) ) {	// make sure a REC_COD exists with FLG_GTO set
				bp += 5;
				dw = GetNum(&bp);	// get address at which the GTO exists
				dw = (dw<<16)|0x0000FFFF;
				if( NULL==(pRec=FindRecAddress(dw)) ) pRec = AddRec(REC_COD, FLG_GTO, "", dw, NULL, 0);	// find or add a REC_ for this address and set it's FLG_GTO
				else pRec->flags |= FLG_GTO;
			}
		} else {				// else normal code line
			if( isdigit(*bp) ) {	// line with address at start (code line)
				dw = GetNum(&bp);	// get address
				lastAddress = (dw<<16) | 0xFFFF;	// combined address with 0xFFFF in lower WORD meaning REAL code, not COMMENT

				while( *bp && *bp++!=':' ) ;	// skip past object codes and (hopefully) ':'
				if( *bp && *bp!=' ' && *bp!='\t' ) {	// if LABEL exists
					for(i=0; i<8 && *bp && *bp!=' ' && *bp!='\t'; ) label[i++] = *bp++;	// get label
					label[i] = 0;
				} else label[0] = 0;
				SkipWhite(bp);	// skip spacing between label and opcode

				// check for and get end-of-line comment
				comment[0] = 0;
				while( *bp ) {
					while( *bp && *bp!='!' ) ++bp;
					if( *bp=='!' ) {
						if( *(bp-1)==' ' || *(bp-1)=='\t' ) {	// found eol comment, save it
							++bp;
							if( *bp==' ' || *bp=='\t' ) ++bp;
							for(i=0; i<sizeof(comment)-1 && *bp; ) comment[i++] = *bp++;
							comment[i] = 0;
						} else ++bp;
					}
				}
				if( NULL==(pRec=FindRecAddress(lastAddress)) ) {
					if( label[0] || comment[0] ) {
						AddRec(REC_UND, 0, label, lastAddress, comment, 0);
					}
				} else {	// rec already created, must have been from a DOT-COMMAND, just add a few things to it
					if( label[0] ) strcpy((char *)pRec->label, label);
					if( comment[0] ) {
						if( pRec->Comment ) free(pRec->Comment);	// if DOT-COMMAND gave this line a comment, let the (possibly) edited comment overwrite it
						pRec->Comment = (char *)malloc(strlen(comment)+1);
						if( pRec->Comment ) strcpy(pRec->Comment, comment);
					}
				}
			} else {	// no dot command, no code address, must be a comment line, DAD, or EQU
				SkipWhite(bp);
				if( *bp==':' ) {	// DAD or EQU
					++bp;	// skip colon
					SkipWhite(bp);
					if( *bp && *bp!=' ' && *bp!='\t' ) {	// if LABEL exists
						for(i=0; i<8 && *bp && *bp!=' ' && *bp!='\t'; ) label[i++] = *bp++;	// get label
						label[i] = 0;
						SkipWhite(bp);
						if( !strnicmp(bp, "DAD", 3) ) {
							bp += 3;
							SkipWhite(bp);
							dw = (GetNum(&bp)<<16) | (0x00008000 + dadcnt++);
							SkipWhite(bp);
							if( *bp!=':' ) {
								printf("ERROR: DAD option ROM# missing on line %d", linenum);
								cnt = 0;
							} else {
								++bp;
								cnt = GetNum(&bp);
							}
							SkipWhite(bp);
							if( *bp=='!' ) {
								++bp;	// skip the !
								if( *bp==' ' || *bp=='\t' ) ++bp;  // skip ONE whitespace, to allow 'indenting' of eol comments
								for(i=0; i<255 && *bp; ) comment[i++] = *bp++;
								if( i==0 ) comment[i++] = ' ';	// make sure comment is not empty
								comment[i] = 0;
							} else comment[0] = 0;
							AddRec(REC_DAD, 0, label, dw, comment[0]==0?NULL:comment, cnt);
						} else if( !strnicmp(bp, "EQU", 3) ) {
							bp += 3;
							SkipWhite(bp);
							dw = (GetNum(&bp)<<16) | (0x0000C000 + equcnt++);
							SkipWhite(bp);
							if( *bp=='!' ) {
								++bp;	// skip the !
								if( *bp==' ' || *bp=='\t' ) ++bp;  // skip ONE whitespace, to allow 'indenting' of eol comments
								for(i=0; i<255 && *bp; ) comment[i++] = *bp++;
								if( i==0 ) comment[i++] = ' ';	// make sure comment is not empty
								comment[i] = 0;
							} else comment[0] = 0;
							pRec = AddRec(REC_EQU, FLG_EQU_DEFINED, label, dw, comment[0]==0?NULL:comment, 0);
						} else printf("ERROR on line %d: something bogus", linenum+1);
					} // if colon but no label, bogus, ignore
				} else {	// full-line comment
					if( *bp=='!' ) ++bp;
					if( *bp==' ' || *bp=='\t' ) ++bp;	// skip possible first WhiteSpace after '!'
					if( *bp==0 ) strcpy(comment, " ");	// blank line, make it a comment with a space
					else {
						for(i=0; i<sizeof(comment)-1 && *bp; ) comment[i++] = *bp++;
						comment[i] = 0;
					}
					// create a record with this full-line comment
					if( lastAddress==0 ) lastAddress = (BaseAddress<<16);
					else if( (lastAddress & 0xFFFF)==0xFFFF ) {	// if first full-line comment after a line of real code
						for(i=linenum+1; i<LstLinesCnt && !isdigit(pLstLines[i][0]); i++) ;	// find next line with real code
						if( i<LstLinesCnt ) {
							bp = pLstLines[i];
							dw = GetNum(&bp);
							lastAddress = (dw<<16);
						} else lastAddress = (lastAddress & 0xFFFF0000)+0x00010000;
					} else ++lastAddress;	// Not first full-line comment in this block, just bump the line# up
					AddRec(REC_REM, 0, "", lastAddress, comment, lastAddress & 0x00007FFF);
				}
			}
		}
	}
}

//*****************************************************************
// Adds DOT-COMMAND lines to LinBuf when creating initial .LST file
void AddLst(char **ppL, char *str) {

	int		len;
	char	*pL;

	pL = *ppL;
	len = strlen(str);
	if( (pL-LstBuf)+len < LstBufLen ) {
		strcpy(pL, str);
		pL += len;
		*pL++ = '\r';
		*pL++ = '\n';
		*ppL = pL;
	}
}

/********************************
 * SYNTAX: HP80DIS <input_name>	*
 ********************************/
int main(int argc, char **argv) {

	int     	i, j, k, g, fullLineCnt=0;
	int     	fp;
	DISASMREC	*pRec;
	DWORD		dw, rp, tp, rep, pp, lastFullLineAddr;
	WORD		addrROMJSB, addrERROR, addrERRORplus;
	char		*pL, label[10], *bp;
	BOOL		gotMODEL=FALSE, gotBASE=FALSE, gotRB=FALSE;

#if DEVTIME
if( argc==1 ) {
//	strcpy(BinFileName, "romsys00000");
	strcpy(BinFileName, "RACEB");
	strcpy(ListFileName, BinFileName);
	printf("%s\r\n", TitleRevStr);
	IsROM = FALSE;
	IsSystemROM = FALSE;
	Is85 = FALSE;
//	IsSystemROM = TRUE;
//	Is85 = FALSE;
	RomPos = 0;
	BaseAddress = 0;//8192*RomPos;	// 0, 20000, 40000, or 60000 octal
	EndAddress = BaseAddress+8192;
} else {
#endif
	// check input arguments and optionally print syntax
	RomPos = 3;
	BaseAddress = 060000;
	EndAddress = BaseAddress+8192;
	for(i=1; i<argc; i++) {
		if( argv[i][0]=='-' || argv[i][0]=='/' ) {
			switch( argv[i][1] ) {
			 case '8':	// 85 or 87, hopefully
			 	if( !gotMODEL ) {
					if( argv[i][2]=='5' ) Is85 = TRUE;
					else if( argv[i][2]=='7' ) Is85 = FALSE;
					else goto defopt;
					gotMODEL = TRUE;
			 	} else printf("ERROR: multiple MODEL options %s", argv[i]);
				break;
			 case 'a':	// ROM 'page' 1,2,3,4 (=addresses, 00000, 20000, 40000, 60000)
			 	if( !gotBASE ) {
					RomPos = atoi(argv[i]+2);
					if( RomPos<1 || RomPos>4 ) goto defopt;
					--RomPos;
					IsROM = TRUE;
					if( RomPos < 3 ) IsSystemROM = TRUE; // if one of the four 'base' System ROMs
					BaseAddress = 8192*RomPos;	// 0, 20000, 40000, or 60000 octal
					EndAddress = BaseAddress+8192;
					gotBASE = TRUE;
			 	} else printf("ERROR: multiple BASE ADDRESS options %s", argv[i]);
			 	break;
			 case 'r':	// force ROM (default)
			 	if( !gotRB ) {
					IsROM = TRUE;
					gotRB = TRUE;
			 	} else printf("ERROR: multiple R/B options %s", argv[i]);
				break;
			 case 'b':	// force BPGM
			 	if( !gotRB ) {
					IsROM = FALSE;
					gotRB = TRUE;
			 	} else printf("ERROR: multiple R/B options %s", argv[i]);
				break;
			 case 's':	// output SOURCE ONLY
			 	OnlySrc = TRUE;
				break;
			 case 'c':	// force end-of-line COMMENT on every line (ie, add a '!')
			 	ForceComment = TRUE;
				break;
			 default:
defopt:
				printf("ERROR: unknown or invalid option: %s\r\n", argv[i]);
defnam:
		printf("\r\n\
%s - HP Series 80 - Disassemble, comment, label, repeat\r\n\
  HP80DIS <Series_80_machine_code_file> [options]\r\n\
    options =\r\n\
       -85 | -87               select appropriate GLOBALS file\r\n\
       -a1 | -a2 | -a3 | -a4   for ROMs, select the address base\r\n\
       -r  | -b                force ROM or BPGM disassembly\r\n\
       -s                      output ONLY source code\r\n\
       -c                      add '!' to end of uncommented lines\r\n\
", TitleRevStr);
				exit(1);
			}
		} else if( BinFileName[0]==0 ) {
			strcpy(BinFileName, argv[i]);
			strcpy(ListFileName, argv[i]);
		} else {
			printf("ERROR: too many filenames: %s", argv[i]);
			exit(1);
		}
	}
#if DEVTIME
}
#endif
	if( BinFileName[0]==0 ) goto defnam;

	//======================================================================
	// LOAD ROM/BINARY PROGRAM
	// open ROM/BPGM file for reading, allocate buffer, read the file in
	if( -1==(fp=open(BinFileName, _O_BINARY | _O_RDONLY)) ) {
		printf("Error reading %s\r\n", BinFileName);
		exit(5);
	}
	InBufLen = lseek(fp, 0, SEEK_END);
	InBuf = (unsigned char *)malloc(InBufLen);
	if( InBuf == NULL ) {
		printf("Error allocating input buffer\r\n");
		exit(6);
	}
	lseek(fp, 0, SEEK_SET);
	read(fp, (void*)InBuf, InBufLen);
	close(fp);

	//======================================================================
	// ROM/BINARY PROGRAM SETUP
	if( IsROM ) {
		curROM = (int)(*InBuf);
		if( RomPos==3 && curROM==0 ) IsSystemROM = TRUE; // if system "plug-in" rom 0
	} else {
//GEEKDO: handle ABS binary programs, and any diffs between 85/87
		// see if has LIFUTIL header on it, if so, remove it
		if( *((WORD*)InBuf)==0x0080 && !strncmp((char*)InBuf+2, "HFSLIF", 6) ) {	// LIF disk
			InBufLen = 512 + *((WORD*)(InBuf+256+28));	// clip end for actual length of file
			memcpy(InBuf, InBuf+512, InBufLen-512);
			InBufLen -= 512;
		}
		if( InBuf[6]!=2 ) {	// if not BPGM file type in Program Control Block (PCB)
			printf("Error: doesn't look like a binary program!\r\n");
			exit(6);
		}
		for(i=7, j=0; i<24; i++) j += (int)InBuf[i];	// rest of PCB 0s?
		if( Is85 && j!=0 ) {
			printf("Error: doesn't look like an HP-85 binary program!\r\n");
			exit(6);
		}
		if( !Is85 && j==0 ) {
			printf("Error: doesn't look like an HP-87 binary program!\r\n");
			exit(6);
		}
		BaseAddress = *((WORD*)(InBuf+24)); // last address BPGM was loaded at
		if( BaseAddress==0 ) BaseAddress = *((WORD*)(InBuf+18));	// if no last loaded address, use ABS BASE ADDRESS
		EndAddress = BaseAddress + InBufLen;
	}

	//======================================================================
	// LOAD and PROCESS GLOBALS FILE
	// open GLOBALS.TXT file (if it exists) for reading, allocate buffer, read the file in
	if( -1!=(fp=open(Is85?"globals85.txt":"globals87.txt", _O_BINARY | _O_RDONLY)) ) {
		LstBufLen = _lseek(fp, 0, SEEK_END);
		// we allocate 1 more than the size of the file so that there's room for
		// a NUL byte at the end in case the idiot user has edited the file and
		// removed the eol seqquence from the end of the last line
		LstBuf = (char *)malloc(LstBufLen+1);
		if( LstBuf == NULL ) {
			printf("Error allocating GLOBALS buffer\r\n");
			exit(8);
		}
		lseek(fp, 0, SEEK_SET);
		read(fp, (void*)LstBuf, LstBufLen);
		close(fp);

// process GLOBALS file, create GLOBAL entries for the DADs and EQUs
		ProcessGLOBALS();
		free(LstBuf);
		LstBuf = NULL;
	}
	LstBufLen = 0;

	//======================================================================
	// LOAD and PROCESS PRE-EXISTING .LST FILE
	// modify input file name to create .LST output file name
	for(i=strlen(ListFileName)-1; i && ListFileName[i]!='.'; i--) ;
	if( i>0 && ListFileName[i]=='.' ) strcpy(ListFileName+i, ".lst");
	else strcat(ListFileName, ".lst");

	// open LST file (if it exists) for reading, allocate buffer, read the file in
	if( -1==(fp=open(ListFileName, _O_BINARY | _O_RDONLY)) ) {	// .LST does NOT exist
		LstBufLen = 2*65536;	// 128K should be WILDLY enough space for preliminary DOT-COMMANDS
		LstBuf = (char *)malloc(LstBufLen);
		if( LstBuf == NULL ) {
			printf("Error allocating LST buffer\r\n");
			exit(7);
		}
		if( BaseAddress==060000 ) {	// if ROM @ 60000 and no .LST file, do some automatic DOT-COMMANDS
			pL = LstBuf;
			// first handle ROM tables
			AddLst(&pL, ".byt 60000 2");
			AddLst(&pL, ".eol 60000 ROM# and complement");
			AddLst(&pL, ".def 60002");
			AddLst(&pL, ".eol 60002 Address of table of runtime routines addresses");
			dw = *((WORD*)(InBuf+2));
			sprintf(ScratBuf, ".lab %o RUNTIM", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".def %o", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			AddLst(&pL, ".eol 60004 Address of table of TOKENS (ASCII keywords)");
			dw = *((WORD*)(InBuf+4));
			sprintf(ScratBuf, ".lab %o TOKENS", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".asp %o", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			dw = *((WORD*)(InBuf+6));
			sprintf(ScratBuf, ".lab %o PARSE", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".def %o", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			AddLst(&pL, ".eol 60006 Address of table of parse routines addresses");
			dw = *((WORD*)(InBuf+8));
			sprintf(ScratBuf, ".lab %o ERMSGS", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			AddLst(&pL, ".eol 60010 Address of table of warning and error messages");
			dw = *((WORD*)(InBuf+10));
			sprintf(ScratBuf, ".lab %o INIT", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			AddLst(&pL, ".eol 60012 Address of ROM initialization routine");

			// second look for known code/non-code sequences like calls to ROMJSB, ERROR, ERROR+
			// first look in GLOBALSxx.TXT for the correct address values for the routines (85/87)
			addrROMJSB = addrERROR = addrERRORplus = 0xFFFF;
			for(i=0; i<GlobalsCnt; i++) {
				if( Globals[i].type==GLO_DAD ) {
					if( !stricmp(Globals[i].label, "ROMJSB") ) addrROMJSB = Globals[i].address;
					if( !stricmp(Globals[i].label, "ERROR") ) addrERROR = Globals[i].address;
					if( !stricmp(Globals[i].label, "ERROR+") ) addrERRORplus = Globals[i].address;
				}
			}
			for(i=0; i<InBufLen-3; i++) {
				if( InBuf[i]==0316 && *((WORD*)(InBuf+i+1))==addrROMJSB ) {	// if JSB =ROMJSB
					dw = BaseAddress+i;
					sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					dw += 3;
					g = InBuf[i+5]==0;	// if target ROM is 0, tell .def to use GLOBALS file
					sprintf(ScratBuf, ".def %o %d", (unsigned int)dw, g);
					AddLst(&pL, ScratBuf);
					dw += 2;
					sprintf(ScratBuf, ".byt %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					dw += 1;
					sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
				}
				if( InBuf[i]==0316 && (*((WORD*)(InBuf+i+1))==addrERROR || *((WORD*)(InBuf+i+1))==addrERRORplus) ) {	// if JSB =ERROR or JSB =ERROR+
					dw = BaseAddress+i;
					sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					dw += 3;
					sprintf(ScratBuf, ".byt %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					dw += 1;
					sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
				}
			}
			// third, handle all RUNTIME routine entry points and primary attribute in front of them
			// Keeping track of processed addresses, since sometimes the same routine will be used for
			// different tokens, and we don't want to duplicate DOT-COMMANDs.
			// Use first 1024 bytes of ScratBuf as bit-flags for 8K possible addresses.

			memset(ScratBuf, 0, 1024);	// set address flags to "not done"
			rp = *((WORD*)(InBuf+2))+2-BaseAddress;	// point to RUNTIM table, skip token 0 which is never used
			tp = (*((WORD*)(InBuf+4)))-BaseAddress;	// point to TOKENS table
			pp = *((WORD*)(InBuf+6))+2;	// point to PARSE table, skip token 0 which is never used
			while( InBuf[tp] != 0377 ) {	// while not end of table
				sprintf(ScratBuf+1024, ".eol %o ", (unsigned int)(rp+BaseAddress));
				j = 1024+strlen(ScratBuf+1024);
				if( (InBuf[tp] & 0x80) && (InBuf[tp] & 0x7F)<' ' ) {	// 87 ASSEMBLER ROM, at least, has some '200' tokens that are bogus
					++tp;	// skip it
					ScratBuf[1024] = 0;	// set buffer to empty
					j = 1024;	// and buffer length
				} else {
					do ScratBuf[j++] = InBuf[tp] & 0x7F;	// copy TOKEN name to RUNTIM entry comment
					while( !(InBuf[tp++] & 0x080) );
					// reached end of this keyword
					ScratBuf[j] = 0;
					AddLst(&pL, ScratBuf+1024);
				}
				// now add DOT-COMMANDs for RUNTIME routine entry point
				dw = *((WORD*)(InBuf+rp));	// get RUNTIME routine entry point address
				if( dw>BaseAddress && dw<BaseAddress+InBufLen ) {
					rp += 2;	// move RUNTIME table ptr to next entry for next loop
					rep = dw-BaseAddress;
					k = 1;	// for primary attribute
					if( (InBuf[rep-1] & 0x0C0)==0 ) k += 1 + ((InBuf[rep-2]>>4)-2+3)/4;	// it's a FUNCTION, include number of SECONDARY ATTRIBUTES!
					else { // statement, add comment to parse table entry
						sprintf(ScratBuf+1024, ".eol %o", (unsigned int)pp);
						ScratBuf[1024+10] = ' ';	// got overwritten by the terminating NUL of the sprintf above
						AddLst(&pL, ScratBuf+1024);
					}
					pp += 2;
					if( !(ScratBuf[rep/8] & (1<<(rep&7))) ) {	// if not processed this one before
						sprintf(ScratBuf+1024, ".cod %o", (unsigned int)dw);	// the entry point has to be code
						AddLst(&pL, ScratBuf+1024);
						sprintf(ScratBuf+1024, ".byt %o %o", (unsigned int)dw-k, k);	// the byte before RUNTIM entry has to be BYT (keyword attribute)
						AddLst(&pL, ScratBuf+1024);
						ScratBuf[rep/8] |= (1<<(rep&7));	// set flag that we've processed this address
					}
				} else {
					rp += 2;
					pp += 2;
				}
			}
			for(i=077773; InBuf[i-BaseAddress]==0; i--) ;	// find how many empty bytes at end of ROM
			++i;
			j = 077774-i;
			if( j ) {
				sprintf(ScratBuf, ".bsz %o %o", i, j);
				AddLst(&pL, ScratBuf);	// make end of ROM a BSZ
			}
			AddLst(&pL, ".byt 77774 4");
			AddLst(&pL, ".eol 77774 ROM checksums");
			LstBufLen = pL-LstBuf;
		} else if( IsROM==FALSE ) {	// BPGM
			pL = LstBuf;
			AddLst(&pL, ".def 32");
			AddLst(&pL, ".eol 32 Address of table of runtime routines addresses");
			dw = *((WORD*)(InBuf+032));
			sprintf(ScratBuf, ".lab %o RUNTIM", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".def %o", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			AddLst(&pL, ".eol 34 Address of table of TOKENS (ASCII keywords)");
			dw = *((WORD*)(InBuf+034));
			sprintf(ScratBuf, ".lab %o TOKENS", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".asp %o", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			dw = *((WORD*)(InBuf+036));
			sprintf(ScratBuf, ".lab %o PARSE", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".def %o", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			AddLst(&pL, ".eol 36 Address of table of parse routines addresses");
			dw = *((WORD*)(InBuf+040));
			sprintf(ScratBuf, ".lab %o ERMSGS", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			AddLst(&pL, ".eol 40 Address of table of warning and error messages");
			dw = *((WORD*)(InBuf+042));
			sprintf(ScratBuf, ".lab %o INIT", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
			AddLst(&pL, ScratBuf);
			AddLst(&pL, ".eol 42 Address of ROM initialization routine");

			// second look for known code/non-code sequences like calls to ROMJSB, ERROR, ERROR+
			// first look in GLOBALSxx.TXT for the correct address values for the routines (85/87)
			addrROMJSB = addrERROR = addrERRORplus = 0xFFFF;
			for(i=0; i<GlobalsCnt; i++) {
				if( Globals[i].type==GLO_DAD ) {
					if( !stricmp(Globals[i].label, "ROMJSB") ) addrROMJSB = Globals[i].address;
					if( !stricmp(Globals[i].label, "ERROR") ) addrERROR = Globals[i].address;
					if( !stricmp(Globals[i].label, "ERROR+") ) addrERRORplus = Globals[i].address;
				}
			}
			for(i=0; i<InBufLen-3; i++) {
				if( InBuf[i]==0316 && *((WORD*)(InBuf+i+1))==addrROMJSB ) {	// if JSB =ROMJSB
					dw = BaseAddress+i;
					sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					dw += 3;
					g = InBuf[i+5]==0;	// if target ROM is 0, tell .def to use GLOBALS file
					sprintf(ScratBuf, ".def %o %d", (unsigned int)dw, g);
					AddLst(&pL, ScratBuf);
					dw += 2;
					sprintf(ScratBuf, ".byt %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					dw += 1;
					sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
				}
				if( InBuf[i]==0316 && (*((WORD*)(InBuf+i+1))==addrERROR || *((WORD*)(InBuf+i+1))==addrERRORplus) ) {	// if JSB =ERROR or JSB =ERROR+
					dw = BaseAddress+i;
					sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					dw += 3;
					sprintf(ScratBuf, ".byt %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
					dw += 1;
					sprintf(ScratBuf, ".cod %o", (unsigned int)dw);
					AddLst(&pL, ScratBuf);
				}
			}
			// third, handle all RUNTIME routine entry points and primary attribute in front of them
			// Keeping track of processed addresses, since sometimes the same routine will be used for
			// different tokens, and we don't want to duplicate DOT-COMMANDs.
			// Use first 4096 bytes of ScratBuf as bit-flags for 32K possible addresses.

			memset(ScratBuf, 0, 4096);	// set address flags to "not done"
			rp = *((WORD*)(InBuf+032))+2-BaseAddress;	// point to RUNTIM table, skip token 0 which is never used
			tp = (*((WORD*)(InBuf+034)))-BaseAddress;	// point to TOKENS table
			pp = *((WORD*)(InBuf+036))+2;	// point to PARSE table, skip token 0 which is never used
			while( InBuf[tp] != 0377 ) {	// while not end of table
				sprintf(ScratBuf+4096, ".eol %o ", (unsigned int)(rp+BaseAddress));
				j = 4096+strlen(ScratBuf+4096);
				if( (InBuf[tp] & 0x80) && (InBuf[tp] & 0x7F)<' ' ) {	// 87 ASSEMBLER ROM, at least, has some '200' tokens that are bogus
					++tp;	// skip it
					ScratBuf[4096] = 0;	// set buffer to empty
					j = 4096;	// and buffer length
				} else {
					do ScratBuf[j++] = InBuf[tp] & 0x7F;	// copy TOKEN name to RUNTIM entry comment
					while( !(InBuf[tp++] & 0x080) );
					// reached end of this keyword
					ScratBuf[j] = 0;
					AddLst(&pL, ScratBuf+4096);
				}
				// now add DOT-COMMANDs for RUNTIME routine entry point
				dw = *((WORD*)(InBuf+rp));	// get RUNTIME routine entry point address
				if( dw>BaseAddress && dw<BaseAddress+InBufLen ) {
					rp += 2;	// move RUNTIME table ptr to next entry for next loop
					rep = dw-BaseAddress;
					k = 1;	// for primary attribute
					if( (InBuf[rep-1] & 0x0C0)==0 ) k += 1 + ((InBuf[rep-2]>>4)-2+3)/4;	// it's a FUNCTION, include number of SECONDARY ATTRIBUTES!
					else { // statement, add comment to parse table entry
						sprintf(ScratBuf+4096, ".eol %o", (unsigned int)pp);
						ScratBuf[4096+10] = ' ';	// got overwritten by the terminating NUL of the sprintf above
						AddLst(&pL, ScratBuf+4096);
					}
					pp += 2;
					if( !(ScratBuf[rep/8] & (1<<(rep&7))) ) {	// if not processed this one before
						sprintf(ScratBuf+4096, ".cod %o", (unsigned int)dw);	// the entry point has to be code
						AddLst(&pL, ScratBuf+4096);
						sprintf(ScratBuf+4096, ".byt %o %o", (unsigned int)dw-k, k);	// the byte before RUNTIM entry has to be BYT (keyword attribute)
						AddLst(&pL, ScratBuf+4096);
						ScratBuf[rep/8] |= (1<<(rep&7));	// set flag that we've processed this address
					}
				} else {
					rp += 2;
					pp += 2;
				}
			}
			LstBufLen = pL-LstBuf;
		} else {
			LstBufLen = 0;
		}
	} else {	// .LST already exists
		LstBufLen = _lseek(fp, 0, SEEK_END);
		// we allocate 1 more than the size of the file so that there's room for
		// a NUL byte at the end in case the idiot user has edited the file and
		// removed the eol seqquence from the end of the last line
		LstBuf = (char *)malloc(LstBufLen+1);
		if( LstBuf == NULL ) {
			printf("Error allocating LST buffer\r\n");
			exit(7);
		}
		lseek(fp, 0, SEEK_SET);
		read(fp, (void*)LstBuf, LstBufLen);
		close(fp);
	}

// process .LST file, create DISASMRECs for comments, labels, data types
	ProcessLST();
	free(LstBuf);
	LstBuf = NULL;
	LstBufLen = 0;

	//======================================================================
	// CREATE TEMP OUTPUT FILE and WRITE INITIAL LINES
	// create/truncate temp .LST file
	if( -1==(OutFile=open(TmpFileName, _O_BINARY | _O_RDWR | _O_CREAT | _O_TRUNC, _S_IREAD | _S_IWRITE)) ) {
		printf("Error creating temp .LST file %s", TmpFileName);
		exit(9);
	}
	sprintf(ScratBuf, ".dis %s %s disassembly - %s", TitleRevStr, BinFileName, TimeString());
	WriteLine(ScratBuf);
	WriteLine(".dis =========================================================================");
	WriteLine(".dis The HP80DIS program assumes that:");
	WriteLine(".dis   1) WhiteSpace at the start of lines is discardable.");
	WriteLine(".dis   2) If the first character on a line is a '.' the line is a DOT COMMAND.");
	WriteLine(".dis   3) If the first character is a ':' is a DAD or EQU.");
	WriteLine(".dis   4) Any line starting with a digit is a line of object/source code.");
	WriteLine(".dis   3) All other lines are full-line comments.  If the first non-WhiteSpace");
	WriteLine(".dis      character is not a '!' then a '!' is added.  Full-line comments are");
	WriteLine(".dis      automatically indented to the normal 'label' column.");
	WriteLine(".dis   4) For HP-85, code LABELS are 1 to 6 characters long and begin in");
	WriteLine(".dis      column 25 (the column immediately after the ':' column).  For HP-87,");
	WriteLine(".dis      code LABELS are 1 to 8 characters long.");
	WriteLine(".dis   5) Actual code starts after WhiteSpace after any LABEL (or non-label).");
	WriteLine(".dis   6) End-of-line comments start after the first '!' character that has a");
	WriteLine(".dis       WhiteSpace character BEFORE it, AFTER the source code.");
	WriteLine(".dis   7) All DADs and EQUs are moved to the start of the file, and there can");
	WriteLine(".dis      be NO full-line comments mixed in with them.  Sorry about that.");
	WriteLine(".dis DOT ('.') COMMANDS can be placed anywhere, each on its own line.");
	WriteLine(".dis ");
	WriteLine(".dis NOTE: All .dis lines are discarded, although this 'help' block is");
	WriteLine(".dis   reinserted each time (so deleting it won't work).  All other");
	WriteLine(".dis   DOT-COMMANDS are moved to the end of the file during processing and are");
	WriteLine(".dis   sorted in increasing address order.");
	WriteLine(".dis =========================================================================");
	WriteLine(".dis");	// to separate above from list of DADs and EQUs

	// Dump all the DADs up front in the same order in which they were originally entered
	j = 0;
	do {
		for(pRec=pDis[0]; pRec!=NULL; pRec = pRec->pNext) {
			if( pRec->type==REC_DAD && (pRec->address & 0x00003FFF)==j ) {
				strcpy(ScratBuf, "                       :");
				strcpy(label, (char*)pRec->label);
				for(i=0; label[i]; i++) ;
				for(; i<8; i++) label[i] = ' ';
				label[i] = 0;
				sprintf(ScratBuf+strlen(ScratBuf), "%s DAD %o:%o", label, (unsigned int)HP85ADDR(pRec->address), pRec->datover);	// option ROM DAD
				if( pRec->Comment ) {
					k = strlen(ScratBuf);
					for(bp=ScratBuf+k; k<51; k++) *bp++ = ' ';
					if( *(bp-1)!=' ' ) *bp++ = ' ';
					*bp++ = '!';
					*bp++ = ' ';
					strcpy(bp, pRec->Comment);
				}
				WriteLine(ScratBuf);
				++j;
				break;
			}
		}
	} while( pRec!=NULL );
	WriteLine(".dis");
	// Dump all the EQUs up front
	j = 0;
	do {
		for(pRec=pDis[0]; pRec!=NULL; pRec = pRec->pNext) {
			if( pRec->type==REC_EQU && (pRec->address & 0x00003FFF)==j ) {
				strcpy(ScratBuf, "                       :");
				strcpy(label, (char*)pRec->label);
				for(i=0; label[i]; i++) ;
				for(; i<8; i++) label[i] = ' ';
				label[i] = 0;
				sprintf(ScratBuf+strlen(ScratBuf), "%s EQU %o", label, (unsigned int)HP85ADDR(pRec->address));
				if( pRec->Comment ) {
					k = strlen(ScratBuf);
					for(bp=ScratBuf+k; k<51; k++) *bp++ = ' ';
					if( *(bp-1)!=' ' ) *bp++ = ' ';
					*bp++ = '!';
					*bp++ = ' ';
					strcpy(bp, pRec->Comment);
				}
				WriteLine(ScratBuf);
				++j;
				break;
			}
		}
	} while( pRec!=NULL );
	WriteLine(".dis");

	if( IsROM ) {
		sprintf(ScratBuf, ".dis                             ABS %o", (unsigned int)BaseAddress);
		WriteLine(ScratBuf);
		NxtByt = (0 + BaseAddress)<<16;
	} else if( Is85 ) {
		for(i=0; i<6; i++) ScratBuf[1024+i] = InBuf[i];
		while( i && ScratBuf[1024+i-1]==' ' ) --i;
		ScratBuf[1024+i+1] = 0;
		sprintf(ScratBuf, ".dis                             NAM %s", (char*)ScratBuf+1024);
		WriteLine(ScratBuf);
		NxtByt = (26 + BaseAddress)<<16;
	} else {
		for(i=0; i<4; i++) ScratBuf[1024+i] = InBuf[i];
		while( i && ScratBuf[1024+i-1]==' ' ) --i;
		ScratBuf[1024+i+1] = 0;
		j = (int)(WORD)InBuf[7];
		sprintf(ScratBuf, ".dis                             NAM %o,%s", j, (char*)ScratBuf+1024);
		WriteLine(ScratBuf);
		NxtByt = (26 + BaseAddress)<<16;
	}

	//======================================================================
	// DISASSEMBLE INPUT FILE TO TEMP FILE
	NextRec = pDis[0];
	do {
/*
if( HP85ADDR(NxtByt)==060562 ) {
	dw =1;
}
*/
		DisAsm();
		WriteLine(ScratBuf);
	} while( HP85ADDR(NxtByt)-BaseAddress < InBufLen );

	if( OnlySrc ) WriteLine("                       :         FIN");
	else WriteLine(".dis                   :         FIN");

	// dump DOT-COMMANDS
	lastFullLineAddr = -1;
	for(pRec=pDis[0]; pRec!=NULL; pRec = pRec->pNext) {
		switch( pRec->type ) {
		 case REC_DRP:
		 	sprintf(ScratBuf, ".drp %o", (unsigned int)(pRec->address>>16));
			WriteLine(ScratBuf);
		 	break;
		 case REC_BYT:
		 	sprintf(ScratBuf, ".byt %o %o", (unsigned int)(pRec->address>>16), pRec->datover);
			WriteLine(ScratBuf);
		 	break;
		 case REC_DEF:
		 	sprintf(ScratBuf, ".def %o %d", (unsigned int)(pRec->address>>16), (pRec->flags & FLG_ROMJSB_DEF));
			WriteLine(ScratBuf);
		 	break;
		 case REC_COD:
		 	sprintf(ScratBuf, ".cod %o %o", (unsigned int)(pRec->address>>16), pRec->datover);
			WriteLine(ScratBuf);
		 	break;
		 case REC_ASC:
		 	sprintf(ScratBuf, ".asc %o %o", (unsigned int)(pRec->address>>16), pRec->datover);
			WriteLine(ScratBuf);
		 	break;
		 case REC_ASP:
		 	sprintf(ScratBuf, ".asp %o", (unsigned int)(pRec->address>>16));
			WriteLine(ScratBuf);
		 	break;
		 case REC_BSZ:
		 	sprintf(ScratBuf, ".bsz %o %o", (unsigned int)(pRec->address>>16), pRec->datover);
			WriteLine(ScratBuf);
		 	break;
		 default:
			break;
		}
		if( pRec->type!=REC_DAD && pRec->type!=REC_EQU ) {
			if( pRec->label[0] ) {
				sprintf(ScratBuf, ".lab %o %s", (unsigned int)(pRec->address>>16), pRec->label);
				WriteLine(ScratBuf);
			}
			if( pRec->Comment!=NULL ) {
				if( pRec->type==REC_REM ) {	// full-line comment
					if( (pRec->address>>16) != lastFullLineAddr ) {
						lastFullLineAddr = (pRec->address>>16);
						fullLineCnt = 0;
					}
					sprintf(ScratBuf, ".com %o %o %s", (unsigned int)(pRec->address>>16), fullLineCnt, pRec->Comment);
					WriteLine(ScratBuf);
					++fullLineCnt;
				} else {	// eol comment
					sprintf(ScratBuf, ".eol %o %s", (unsigned int)(pRec->address>>16), pRec->Comment);
					WriteLine(ScratBuf);
				}
			}
		}
		if( pRec->flags & FLG_FEQ ) {
			sprintf(ScratBuf, ".feq %o %s", (unsigned int)HP85ADDR(pRec->address), pRec->reflabel);
			WriteLine(ScratBuf);
		}
		if( pRec->flags & FLG_FDA ) {
			sprintf(ScratBuf, ".fda %o %s", (unsigned int)HP85ADDR(pRec->address), pRec->reflabel);
			WriteLine(ScratBuf);
		}
		if( pRec->flags & FLG_GLO ) {
			sprintf(ScratBuf, ".fgl %o %s", (unsigned int)HP85ADDR(pRec->address), pRec->reflabel);
			WriteLine(ScratBuf);
		}
		if( pRec->flags & FLG_GTO ) {
			sprintf(ScratBuf, ".gto %o", (unsigned int)HP85ADDR(pRec->address));
			WriteLine(ScratBuf);
		}
	}
	close(OutFile);

	//======================================================================
	// DELETE OLD .LST FILE and RENAME TEMP FILE TO .LST FILE
	if( !BailOut ) {
		unlink(ListFileName);
		rename(TmpFileName, ListFileName);
	}

	exit(0);
}

//***************
//* END OF FILE *
//***************
